import numpy as np
from keras.layers import Embedding, Input, Dense, Bidirectional, LSTM, Concatenate, Add, Multiply,Subtract,SpatialDropout1D
from keras.models import Model,load_model




class AccuracyPredictionModelFull:
    
    def __init__(self,main_embedding,alignment_embedding=None,sense_embedding=None,relation_embedding=None):
        self.e = main_embedding
        self.ea = alignment_embedding
        self.es = sense_embedding
        self.er = relation_embedding
    
    def return_instance(self,LSTMLAYERS=2,SENSES=False,POINTERS=False,HIER=True,dim_input=128,dim_lstm=128,out_dim_main=3,out_dim_rest=36):

        in_deps = Input((dim_input,))
        in_amrs = Input((dim_input,))

        in_alignsdep = Input((dim_input,))
        in_alignsamr = Input((dim_input,))

        in_relsdep = Input((dim_input,))
        in_relsamr = Input((dim_input,))

        in_sensesamr = Input((dim_input,))


        lstm_layers_amr = []
        lstm_layers_dep = []

        for i in range(LSTMLAYERS):
            if i == LSTMLAYERS-1:
                lstm_layers_amr.append(Bidirectional(LSTM(dim_lstm,return_sequences=False)))
                lstm_layers_dep.append(Bidirectional(LSTM(dim_lstm,return_sequences=False)))
            else:
                lstm_layers_amr.append(Bidirectional(LSTM(dim_lstm,return_sequences=True)))
                lstm_layers_dep.append(Bidirectional(LSTM(dim_lstm,return_sequences=True)))


        def apply_lstm(lstmlist,X):
            for lstm in lstmlist:
                X=lstm(X)
            return X



        amr_embed = self.e(in_amrs)
        dep_embed = self.e(in_deps)

        alignamr_embed = self.ea(in_alignsamr)
        aligndep_embed = self.ea(in_alignsdep)

        relamr_embed = self.er(in_relsamr)
        reldep_embed = self.er(in_relsdep)

        senses_embed= self.es(in_sensesamr)

        if SENSES and POINTERS:
            ae = Add()([amr_embed,alignamr_embed,relamr_embed,senses_embed])
            de = Add()([dep_embed,aligndep_embed,reldep_embed])

        elif SENSES:
            ae = Add()([amr_embed,relamr_embed,senses_embed])
            de = Add()([dep_embed,reldep_embed])

        elif POINTERS:
            ae = Add()([amr_embed,relamr_embed,senses_embed])
            de = Add()([dep_embed,reldep_embed])

        ae = SpatialDropout1D(0.2)(ae)
        de = SpatialDropout1D(0.2)(de)
        
        ae = apply_lstm(lstm_layers_amr,ae)
        de = apply_lstm(lstm_layers_dep,de)

        shareds = Subtract()([ae,de])
        sharedm = Multiply()([ae,de])
        shareda = Add()([ae,de])


        combo_embed=Concatenate()([shareds,sharedm,shareda])

        out = Dense(out_dim_main,activation="sigmoid")(combo_embed)

        if HIER:
            out_rest=Dense(out_dim_rest,activation="sigmoid")(combo_embed)
            combo2=Concatenate()([out_rest,combo_embed])
            out=Dense(out_dim_main,activation="sigmoid")(combo2)

        if HIER:
            model = Model(inputs=[in_deps, in_alignsdep,in_relsdep,in_amrs,in_alignsamr,in_relsamr,in_sensesamr], outputs=[out,out_rest])
            model.compile(optimizer='adam', loss='mean_squared_error',loss_weights=[1,0.2])
        else:
            model = Model(inputs=[in_deps, in_alignsdep,in_relsdep,in_amrs,in_alignsamr,in_relsamr,in_sensesamr], outputs=[out])
            model.compile(optimizer='adam', loss='mean_squared_error')
        print(model.summary())
        return model
