import pickle
import sys
import re
import json
from amr_helpers import split_simple, padcrop
import numpy as np
from keras.layers import Embedding, Input, Dense, Bidirectional, LSTM, Concatenate, Add, Multiply,Subtract,Flatten,Conv1D,MaxPooling1D,SpatialDropout1D
from keras.models import Model,load_model
from scipy.stats import pearsonr
from sklearn.metrics import mean_squared_error as mse
from operator import itemgetter
from collections import defaultdict
#from keras_self_attention import SeqSelfAttention
from config import CONFIG



def loadj(p):
    with open(p,"r") as f:
        return json.load(f)


def loadjmultiple(paths):
    paths = paths.split(",")

    multiple = {}
    for i,p in enumerate(paths):
        dat = loadj(p)
        for key in dat:
            multiple[p+"-"+key] =dat[key]
    return multiple


def make_word_index(lss):
    dic={"NONE":0,"PAD":1}
    i=2
    for ls in lss:
        for l in ls:
            for word in l:
                if word in dic:
                    continue
                else:
                    dic[word] = i
                    i+=1
    return dic

def padcrop(seqs,length=25):
    seqs=[(length-len(seq))*["PAD"]+seq for seq in seqs]
    seqs = [seq[:length] for seq in seqs]
    return seqs

def maybe_id(x,word_index):
    if x in word_index:
        return word_index[x]
    else:
        return word_index["NONE"]

def idlookup(seqs,word_index):
    seqs = [[maybe_id(x,word_index) for x in seq] for seq in seqs]
    return seqs

def targetdics2array(dics,order=None,only=[]):
    out = []
    if not order:
        order = list([key for key in dics[0]])
        if only:
            order=[x for x in order if x in only]
    for dic in dics:
        tmp=[]
        for key in order:
            #print(key,order,dic)
            tmp+=dic[key]
        out.append(tmp)
    return np.array(out),order

def remove_senses(stringseqs,splitter=" "):
    out1=[]
    out2=[]
    pattern=re.compile(r"[a-z]+-[0-9]+")
    for string in stringseqs:
        tokens=string.split(" ")
        senses=[]
        newts=[]
        for t in tokens:
            if pattern.match(t):
                sense = t.split("-")
                senses.append(sense[1])
                newts.append(sense[0])
            else:
                senses.append("NONE")
                newts.append(t)
        out2.append(" ".join(senses))
        out1.append(" ".join(newts))
    return out1,out2

def clean_deps(stringseqs):
    return [string.replace("-","xhyphen") for string in stringseqs]

def clean_common(stringseqs,splitter=" "):
    out=[]
    for seq in stringseqs:
        tokens=seq.split(splitter)
        for i in range(len(tokens)):
            if "_" in tokens[i]:
                tokens[i] = "SUB"
            if isint(tokens[i]):
                tokens[i] = "INT"
            elif isfloat(tokens[i]):
                tokens[i] = "FLOAT"
            else:
                tokens[i] = tokens[i].lower().replace("\"","").replace("\'","")
        out.append(splitter.join(tokens))
    return out

        
def isfloat(x):
    try:
        a = float(x)
    except ValueError:
        return False
    else:
        return True

def isint(x):
    try:
        a = int(x)
    except ValueError:
        return False
    else:
        return True

class Tokenizer:
    def __init__(self,splitter=" ",pretrained=None,mincount=5,lower=False):
        self.pretrained=pretrained
        self.splitter=splitter
        self.mincount=mincount
        self.lower=lower
        return None

    def fit(self,stringseqs,setzero={}):
        self.wi = self.make_word_index(stringseqs,setzero)
    
    def simple_clean(self,string):
        string=string.replace("\"","").replace("\'","")
        #if self.islindep:
        #    string=string.replace("-","xhyphen")
        return string

    def make_word_index(self,ls,setzero={}):
        dic={"NONE":0,"-PAD-":1,"SOS":2,"EOS":3,"ZERO":4,"INT":5,"FLOAT":6,"NAME":7,"SUB":8}
        i=9
        self.word_count = defaultdict(int)
        for l in ls:
            string=self.simple_clean(l)
            for w in string.split(self.splitter):
                self.word_count[w]+=1
        for l in ls:
            string=self.simple_clean(l)
            for w in string.split(self.splitter):
                if w in dic:
                    continue
                elif any(w.startswith(s) for s in setzero):
                    dic[w] = 4
                elif self.word_count[w]>=self.mincount:
                    dic[w] = i
                    i+=1
                else:
                    dic[w] = 0
        return dic

    def most_common(self,n=50):
        return list(sorted(self.word_count.items(),key=itemgetter(1),reverse=True))[:n]
    
    def least_common(self,n=50):
        return list(sorted(self.word_count.items(),key=itemgetter(1),reverse=False))[:n]

    def toids(self,tokens):
        ls = []
        for t in tokens:
            ls.append(self.wi[t])
        return ls
    
    def padcrop(self,tokens,length):
        while len(tokens) < length:
            tokens=["-PAD-"]+tokens
        return tokens[:length]
    
    def clean_tokens(self,tokens):
        if self.lower:
            tokens=[t.lower() for t in tokens]
        tokens=["SOS"]+tokens
        tokens=tokens+["EOS"]
        for i in range(len(tokens)):
            if tokens[i] not in self.wi:
                tokens[i]="NONE"
        return tokens
                

    
    def to_id_seqs(self,stringseqs,length=128):
        out = []
        for string in stringseqs:
            #print("tokens I",string)
            tokens=self.simple_clean(string).split(self.splitter)
            tokens = self.clean_tokens(tokens)
            #print("tokens II"," ".join(tokens))
            padded = self.padcrop(tokens,length=length)
            #print("tokens III"," ".join(padded))
            ids = self.toids(padded)
            #print("IDS",ids)
            out.append(ids)
        return out

    def make_embedding(self,dim=128):
        if self.pretrained:
            return None
        else:
            return Embedding(len(set(self.wi.values())), dim, trainable=True)

    def save(self,path):
        with open(path,"wb") as f:
            pickle.dump(self,f)

def load_tokenizer(path):
    with open(path,"rb") as f:
        return pickle.load(f)
