import sys
import os
os.environ["CUDA_VISIBLE_DEVICES"]="0"
import re
from model import AccuracyPredictionModelFull
import json
import numpy as np
from scipy.stats import pearsonr
from sklearn.metrics import mean_squared_error as mse
from operator import itemgetter
from collections import defaultdict
from config import CONFIG
import more_helpers as helpers
import pickle


CF=CONFIG[sys.argv[5]]

HIER=CF["HIER"]
SHAREW=CF["SHAREW"]
LSTMLAYERS = CF["LSTMLAYERS"]
POINTERS = CF["POINTERS"]
SENSES=CF["SENSES"]
DEP = CF["DEP"]

        
        

train = helpers.loadjmultiple(sys.argv[1])
dev = helpers.loadjmultiple(sys.argv[2])
test = helpers.loadjmultiple(sys.argv[3])
runid = sys.argv[4]


trainkeys = list(sorted(list(train.keys())))
devkeys = list(sorted(list(dev.keys())))
testkeys = list(sorted(list(test.keys())))


if HIER:
    target_train,order = helpers.targetdics2array([train[k]["eval"] for k in trainkeys],order=None,only=["Smatch "])
    target_dev,_ = helpers.targetdics2array([dev[k]["eval"] for k in devkeys],order=order)
    target_test,_ = helpers.targetdics2array([test[k]["eval"] for k in testkeys],order=order)
    target_train_rest,order_rest = helpers.targetdics2array([train[k]["eval"] for k in trainkeys],order=None,only=[])
    target_dev_rest,_ = helpers.targetdics2array([dev[k]["eval"] for k in devkeys],order=order_rest)
    target_test_rest,_ = helpers.targetdics2array([test[k]["eval"] for k in testkeys],order=order_rest)

else:
    target_train,order = helpers.targetdics2array([train[k]["eval"] for k in trainkeys],order=None,only=[])
    target_dev,_ = helpers.targetdics2array([dev[k]["eval"] for k in devkeys],order=order)
    target_test,_ = helpers.targetdics2array([test[k]["eval"] for k in testkeys],order=order)

with open("models/orderedmetrics_"+runid+".pkl","wb") as f:
    pickle.dump((order,order_rest),f)


trainamrs,trainsensesamr = helpers.remove_senses([train[key]["amr_lin"] for key in trainkeys])
devamrs,devsensesamr = helpers.remove_senses([dev[key]["amr_lin"] for key in devkeys])
testamrs,testsensesamr = helpers.remove_senses([test[key]["amr_lin"] for key in testkeys])

trainamrs = helpers.clean_common(trainamrs)
devamrs = helpers.clean_common(devamrs)
testamrs = helpers.clean_common(testamrs)

traindeps = helpers.clean_common(helpers.clean_deps([train[key]["dep_lin_no_ids"] for key in trainkeys]))
devdeps = helpers.clean_common(helpers.clean_deps([dev[key]["dep_lin_no_ids"] for key in devkeys]))
testdeps = helpers.clean_common(helpers.clean_deps([test[key]["dep_lin_no_ids"] for key in testkeys]))

if not DEP:
    traindeps = helpers.clean_common(helpers.clean_deps([" ".join(train[key]["tokens"]) for key in trainkeys]))
    devdeps = helpers.clean_common(helpers.clean_deps([" ".join(dev[key]["tokens"]) for key in devkeys]))
    testdeps = helpers.clean_common(helpers.clean_deps([" ".join(test[key]["tokens"]) for key in testkeys]))


trainrelsamr = [train[key]["amrrels"] for key in trainkeys]
devrelsamr = [dev[key]["amrrels"] for key in devkeys]
testrelsamr = [test[key]["amrrels"] for key in testkeys]

trainalignamr = [train[key]["amralign"] for key in trainkeys]
devalignamr = [dev[key]["amralign"] for key in devkeys]
testalignamr = [test[key]["amralign"] for key in testkeys]

trainrelsdep = [train[key]["deprels"] for key in trainkeys]
devrelsdep = [dev[key]["deprels"] for key in devkeys]
testrelsdep = [test[key]["deprels"] for key in testkeys]

trainaligndep = [train[key]["depalign"] for key in trainkeys]
devaligndep = [dev[key]["depalign"] for key in devkeys]
testaligndep = [test[key]["depalign"] for key in testkeys]

if not DEP:
    trainaligndep = [" ".join([str(num) for num in list(range(0,len(train[key]["depalign"].split(" "))))]) for key in trainkeys]
    devaligndep = [" ".join([str(num) for num in list(range(0,len(dev[key]["depalign"].split(" "))))]) for key in devkeys]
    testaligndep = [" ".join([str(num) for num in list(range(0,len(test[key]["depalign"].split(" "))))]) for key in testkeys]




#3 embedding spaces: alignments and brackets, words, concepts
to = helpers.Tokenizer(" ")
to.fit(trainamrs+traindeps,setzero={"[","]",":"})
to.save("models/maintokenizer_"+runid+".pkl")

print(to.most_common())
print(to.least_common())
print(len([k for k in to.word_count if to.word_count[k] < 2]))

tor = helpers.Tokenizer(" ")
tor.fit(trainrelsamr+trainrelsdep)
tor.save("models/reltokenizer_"+runid+".pkl")
print(tor.most_common())
print(tor.least_common())

toa = helpers.Tokenizer(" ")
toa.fit(trainalignamr+trainaligndep)
toa.save("models/aligntokenizer_"+runid+".pkl")

print(toa.most_common())
print(toa.least_common())

tos = helpers.Tokenizer(" ")
tos.fit(trainsensesamr)
tos.save("models/sensetokenizer_"+runid+".pkl")
print(tos.most_common())
print(tos.least_common())


traindeps = to.to_id_seqs(traindeps)
devdeps = to.to_id_seqs(devdeps)
testdeps = to.to_id_seqs(testdeps)

trainamrs = to.to_id_seqs(trainamrs)
devamrs = to.to_id_seqs(devamrs)
testamrs = to.to_id_seqs(testamrs)

trainrelsamr = tor.to_id_seqs(trainrelsamr)
devrelsamr = tor.to_id_seqs(devrelsamr)
testrelsamr = tor.to_id_seqs(testrelsamr)

trainrelsdep = tor.to_id_seqs(trainrelsdep)
devrelsdep = tor.to_id_seqs(devrelsdep)
testrelsdep = tor.to_id_seqs(testrelsdep)

trainalignamr = toa.to_id_seqs(trainalignamr)
devalignamr = toa.to_id_seqs(devalignamr)
testalignamr = toa.to_id_seqs(testalignamr)

trainaligndep = toa.to_id_seqs(trainaligndep)
devaligndep = toa.to_id_seqs(devaligndep)
testaligndep = toa.to_id_seqs(testaligndep)

trainsensesamr = tos.to_id_seqs(trainsensesamr)
devsensesamr = tos.to_id_seqs(devsensesamr)
testsensesamr = tos.to_id_seqs(testsensesamr)

print(traindeps[:5])
print(trainamrs[:5])
print(trainrelsdep[:5])
print(trainalignamr[:5])
print(trainaligndep[:5])
print(trainsensesamr[:5])
print(trainrelsamr[:5])

e = to.make_embedding()
ea = toa.make_embedding()
es = tos.make_embedding()
er = tor.make_embedding()

model = AccuracyPredictionModelFull(e,ea,es,er).return_instance(LSTMLAYERS=LSTMLAYERS,SENSES=SENSES,POINTERS=POINTERS,dim_input=128,dim_lstm=128)
print(model.summary())

tmpcc = -1
lowestmse=1
for epoch in range(10):
    print("EPOCH",epoch)
    if HIER:
        model.fit([traindeps,trainaligndep,trainrelsdep,trainamrs,trainalignamr,trainrelsamr,trainsensesamr], [target_train,target_train_rest],
                  epochs=1, verbose=2, batch_size=128)
        dev_preds = model.predict([devdeps,devaligndep,devrelsdep,devamrs,devalignamr,devrelsamr,devsensesamr])[0]
    else:
        model.fit([traindeps,trainaligndep,trainrelsdep,trainamrs,trainalignamr,trainrelsamr,trainsensesamr], [target_train],
                      epochs=1, verbose=2, batch_size=128)
        dev_preds = model.predict([devdeps,devaligndep,devrelsdep,devamrs,devalignamr,devrelsamr,devsensesamr])
    sum_mse = mse(dev_preds.flatten(),target_dev.flatten())
    pr = pearsonr(dev_preds.flatten(),target_dev.flatten())
    print ("eval on dev") 
    print("sum mse:",sum_mse,"peasronr",pr)
    sumpcc = pr[0]
    if sumpcc > tmpcc: 
        print("evaluation score increased from",tmpcc,"to",sumpcc,"by",sumpcc-tmpcc)
        print("saving model to:",'models/'+runid+'-model.h5')
        model.save_weights('models/'+runid+'-model.h5')
        tmpcc = sumpcc
        lowestmse=sum_mse
print("training finished, loading best on val")
model.load_weights('models/'+runid+'-model.h5')
print("evaluating on TEST...")
if HIER:
    testpreds = model.predict([testdeps,testaligndep,testrelsdep,testamrs,testalignamr,testrelsamr,testsensesamr])[0]
else:
    testpreds = model.predict([testdeps,testaligndep,testrelsdep,testamrs,testalignamr,testrelsamr,testsensesamr])
for j,i in enumerate(range(0,len(order)*3,3)):
    print(order[int(i/3)])
    print("PRECISION")
    print("mse:",mse(target_test[:,i],testpreds[:,i]))
    print("pearsonr:",pearsonr(target_test[:,i],testpreds[:,i]))
    print("RECALL")
    print("mse:",mse(target_test[:,i+1],testpreds[:,i+1]))
    print("pearsonr:",pearsonr(target_test[:,i+1],testpreds[:,i+1]))
    print("F1")
    print("mse:",mse(target_test[:,i+2],testpreds[:,i+2]))
    print("pearsonr:",pearsonr(target_test[:,i+2],testpreds[:,i+2]))

print("process finished")
